import numpy as np
from peakutils import peak
import warnings

from .utils import flux_to_mag


class Lightcurve(object):
    """Light curve class."""

    def __init__(self, band, lcs_df):
        """

        Parameters
        ----------
        band: str
            Light-curve filter/band name.
        lcs_df: Dataframe
            Light-curve data: `times`, `fluxes`, `flux_errors`, `magnitudes`, `mag_errors`
            and `magnitude system`.
        """
        self.band = band

        data = lcs_df[lcs_df.band == band]
        self.times = data.time.values
        self.fluxes = data.flux.values
        self.flux_errors = data.flux_err.values
        self.zp = float(data.zp.unique()[0])
        self.magnitudes, self.mag_errors = flux_to_mag(self.fluxes, self.flux_errors, self.zp)
        self.mag_sys = data.mag_sys.unique()[0]

    def __repr__(self):
        return (f"band: {self.band}, zp: {self.zp:.5}, mag_sys: {self.mag_sys}"
                "\nother attributes: times, fluxes, flux_errors, magnitudes, mag_errors")

    def __getitem__(self, item):
        return getattr(self, item)

    def mask_lc(self, mask, copy=False):
        """Masks the light-curve data with the given mask.

        Parameters
        ----------
        mask: bool list
            Mask to apply to the light curve.
        mask: bool, default ``False``
            If ``True``, the masked light curves are saved on
            separate arrays.
        """
        if not copy:
            self.times = self.times[mask]
            self.fluxes = self.fluxes[mask]
            self.flux_errors = self.flux_errors[mask]
            self.magnitudes = self.magnitudes[mask]
            self.mag_errors = self.mag_errors[mask]
        else:
            self.masked_times = self.times[mask]
            self.masked_fluxes = self.fluxes[mask]
            self.masked_flux_errors = self.flux_errors[mask]
            self.masked_magnitudes = self.magnitudes[mask]
            self.masked_mag_errors = self.mag_errors[mask]

    def get_max(self):
        """Calculates the peak magnitude (:math:`m_{max}`) and
        its epoch (:math:`t_{max}`).
        """
        magnitudes = np.nan_to_num(self.magnitudes, nan=np.nanmean(self.magnitudes))
        peak_ids = peak.indexes(-magnitudes, thres=0.3, min_dist=len(self.times) // 3)
        if len(peak_ids) == 0:
            self.mmax = self.mmax_err = np.nan
            self.tmax = np.nan
            self.tmax_err = np.nan
        else:
            self.mmax = self.magnitudes[peak_ids[0]]
            self.mmax_err = self.mag_errors[peak_ids[0]]
            self.tmax = self.times[peak_ids[0]]

            # get tmax_err
            # use only data around peak
            mask = (self.times > self.tmax - 5) & (self.times < self.tmax + 5)
            times = self.times[mask]
            brightest_mag = (self.magnitudes - self.mag_errors)[mask]
            id_err = np.argmin(np.abs(brightest_mag - self.mmax))
            self.tmax_err = np.abs(times[id_err] - self.tmax)

    def get_dm15(self):
        r"""Calculates the classic parameter :math:`\Delta m_{15}`
        (Phillips 1993), but using the epoch of peak magnitude of
        the current light curve as reference (not necessarily :math:`B`
        band).
        """
        self.get_max()
        if np.isnan(self.tmax):
            self.dm15 = self.dm15_err = np.nan
        else:
            phases = self.times - self.tmax
            if any(np.abs(phases - 15) < 0.5):
                dm15_id = np.argmin(np.abs(phases - 15))
                self.dm15 = self.magnitudes[dm15_id] - self.mmax
                self.dm15_err = np.sqrt(self.mag_errors[dm15_id] ** 2 + self.mmax_err**2)
            else:
                self.dm15 = self.dm15_err = np.nan


class Lightcurves(object):
    """Multi-colour light curves class."""

    def __init__(self, lcs_df):
        self.bands = lcs_df.band.unique()

        for band in self.bands:
            lc = Lightcurve(band, lcs_df)
            setattr(self, band, lc)

    def __repr__(self):
        return str(self.bands)

    def __getitem__(self, item):
        return getattr(self, item)

    def get_max_colour(self, band1, band2):
        """Calculates the colour at the epoch of peak magnitude
        in ``band1``.

        The colour is :math:`(band1 - band2)` at the time of
        peak magnitude in ``band1``.

        Parameters
        ----------
        band1: str
            Name of first band.
        band2: str
            Name of second band.

        Returns
        -------
        colour: float
            Colour at the epoch of peak magnitude in ``band1``.
        colour_err: float
            Uncertainty on the colour parameter.
        """
        cond1 = band1 in self.bands
        cond2 = band2 in self.bands
        assert cond1 and cond2, f"band(s) not in {self.bands}"

        self[band1].get_max()
        tmax = self[band1].tmax
        mmax = self[band1].mmax
        mmax_err = self[band1].mmax_err
        if np.isnan(tmax):
            warnings.warn(f"cannot estimate time of peak for {band1}")
            colour = colour_err = np.nan
        else:
            rel_phases = self[band2].times - tmax
            if any(np.abs(rel_phases) < 0.5):
                band2_id = np.argmin(np.abs(rel_phases))
                mag2 = self[band2].magnitudes[band2_id]
                mag2_err = self[band2].mag_errors[band2_id]
                colour = mmax - mag2
                colour_err = np.sqrt(mmax_err**2 + mag2_err**2)
            else:
                warnings.warn(f"{band2} does not have data at time of peak for {band1}")
                colour = colour_err = np.nan

        return colour, colour_err

    def get_colour_stretch(self, band1, band2):
        """Calculates the colour stretch parameter (Burns et al. 2014).

        The formula used is :math:`(T(band1-band2)_{max} - T_{max}/30)`, where
        :math:`T(band1−band2)_{max}` is the time of maximum (reddest colour)
        in the :math:`(band1−band2)` colour curve and :math:`T_{max}` is the
        epoch of peak magnitude in ``band1``.

        Parameters
        ----------
        band1: str
            Name of first band.
        band2: str
            Name of second band.

        Returns
        -------
        stretch: float
            Colour-stretch parameter.
        stretch_err: float
            Uncertainty on the colour-stretch parameter.
        """
        cond1 = band1 in self.bands
        cond2 = band2 in self.bands
        assert cond1 and cond2, f"band(s) not in {self.bands}"

        colour_curve = self[band1].magnitudes - self[band2].magnitudes
        colour_curve_err = np.sqrt(self[band1].mag_errors ** 2 + self[band2].mag_errors ** 2)

        self[band1].get_max()
        tmax = self[band1].tmax
        tmax_err = self[band1].tmax_err
        times = self[band1].times

        if np.isnan(tmax):
            warnings.warn(f"cannot estimate time of peak for {band1}")
            stretch = stretch_err = np.nan
        else:
            mask = times > tmax
            colour_curve = np.copy(colour_curve[mask])
            times = np.copy(times[mask])
            peak_ids = peak.indexes(colour_curve, thres=0.3, min_dist=len(times) // 3)
            if len(peak_ids) > 0:
                colour_tmax = times[peak_ids[0]]
                stretch = (colour_tmax - tmax) / 30
                stretch_err = tmax_err / 30  # inaccurate error propagation
            else:
                # warnings.warn(f"The peak in the colour curve was not found")
                stretch = stretch_err = np.nan

        return stretch, stretch_err

    def get_lc_params(self):
        r"""Calculates the peak magnitude (:math:`m_{max}`), its epoch
        (:math:`t_{max}`) and :math:`\Delta m_{15}` for all bands.
        """
        for band in self.bands:
            self[band].get_max()
            self[band].get_dm15()



import os
import glob
import numpy as np
import matplotlib.pyplot as plt

import warnings
import piscola
from .extinction_correction import extinction_filter

class SingleFilter(object):
    """Single filter class."""

    def __init__(self, band, mag_sys):
        """
        Parameters
        ----------
        band: str
            Name of the band.
        mag_sys: str
            Magnitude system.
        """
        self.name = band
        self.mag_sys = mag_sys
        self._add_filter(band, mag_sys)

    def __repr__(self):
        rep = (
            f"name: {self.name}, eff_wave: {self.eff_wave:.1f} Å,"
            f" response_type: {self.response_type}, mag_sys: {self.mag_sys}"
        )
        return rep

    def __getitem__(self, item):
        return getattr(self, item)

    def _add_filter(self, filt_name, mag_sys):
        """Adds a filter from the available filters
        in the PISCOLA library.

        Parameters
        ----------
        filt_name: str
            Name of the filter.
        mag_sys: str
            Magnitude system.
        """
        pisco_path = piscola.__path__[0]
        filt_pattern = os.path.join(pisco_path, "filters", "*", f"{filt_name}.dat")
        filt_file = glob.glob(filt_pattern, recursive=True)

        err_message = (
            "No filter file or multiple files with "
            f"the pattern {filt_pattern} found."
        )
        assert len(filt_file) == 1, err_message

        filt_file = filt_file[0]
        self.filt_file = filt_file

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            wavelength, transmission = np.loadtxt(filt_file).T
        # remove long tails of zero values on both edges
        imin, imax = trim_filters(transmission)
        self.wavelength = wavelength[imin:imax]
        self.transmission = transmission[imin:imax]

        # retrieve response type; if none, assumed to be photon type
        filt_dir = os.path.dirname(filt_file)
        resp_file = os.path.join(filt_dir, "response_type.txt")
        if os.path.isfile(resp_file):
            with open(resp_file) as file:
                line = file.readlines()[0]
                self.response_type = line.split()[0].lower()
        else:
            self.response_type = "photon"

        err_message = (
            f'"{self.response_type}" is not a valid response type '
            f'"photon" or "energy") for {filt_name} filter.'
        )
        assert self.response_type in ["photon", "energy"], err_message

        self.eff_wave = self.calc_eff_wave()

        readme_file = os.path.join(filt_dir, "README.txt")
        if os.path.isfile(readme_file):
            with open(readme_file, "r") as file:
                self.comments = file.read()
        else:
            self.comments = ""

        self.mag_sys = mag_sys

    def calc_eff_wave(self, sed_wave=None, sed_flux=None):
        """Calculates the effective wavelength.

        The effective wavelength depends on the shape
        of the given SED.

        Parameters
        ----------
        sed_wave : array-like, default ``None``
            SED's wavelength range. If ``None``, a flat SED is used.
        sed_flux : array-like, default ``None``
            SED's flux density distribution. If ``None``, a flat SED is used.

        Returns
        -------
        eff_wave: float
            Effective wavelength.
        """
        if sed_wave is None or sed_flux is None:
            sed_wave = self.wavelength.copy()
            sed_flux = 100 * np.ones_like(sed_wave)

        transmission = self.transmission.copy()
        # check filter response type
        if self.response_type == "energy":
            transmission /= self.wavelength

        transmission = np.interp(sed_wave, self.wavelength, transmission, left=0.0, right=0.0)
        I1 = np.trapezoid((sed_wave**2) * transmission * sed_flux, sed_wave)
        I2 = np.trapezoid(sed_wave * transmission * sed_flux, sed_wave)
        eff_wave = I1 / I2

        return eff_wave

    def integrate_filter(self, sed_wave, sed_flux):
        """Calculates the flux density of an SED given a filter response.

        Parameters
        ----------
        sed_wave : array-like
            SED's wavelength range.
        sed_flux : array-like
            SED's flux density distribution.

        Returns
        -------
        flux_filter : float
            Flux density.
        """

        blue_edge_covered = sed_wave.min() <= self.wavelength.min()
        red_edge_covered = sed_wave.max() >= self.wavelength.max()
        err_message = f"The SED does not completely overlap with {self.name} filter."
        assert blue_edge_covered and red_edge_covered, err_message

        transmission = self.transmission.copy()
        # check filter response type
        if self.response_type == "energy":
            transmission /= self.wavelength

        transmission = np.interp(sed_wave, self.wavelength, transmission, left=0.0, right=0.0)
        I1 = np.trapezoid(sed_flux * transmission * sed_wave, sed_wave)
        I2 = np.trapezoid(self.transmission * self.wavelength, self.wavelength)
        flux_filter = I1 / I2

        return flux_filter

    def _get_standard_flux(self, mag_sys_file):
        """Calculates the integrated flux of a standard star SED.

        Parameters
        ----------
        mag_sys_file: str
            File with the magnitude system information.

        Returns
        -------
        f_sed: float
            Integrated flux density.
        """
        pisco_path = piscola.__path__[0]

        # get standard SED file name
        with open(mag_sys_file, "rt") as file:
            for line in file:
                if len(line) > 0:
                    if "standard_sed:" in line.split():
                        standard_sed = line.split()[1]
                        break

        assert (
            "standard_sed" in locals()
        ), f"Standard SED file not found in {mag_sys_file}"

        if standard_sed.lower() == "ab":
            c = 2.99792458e18  # speed of light in [Å/s]
            sed_wave = np.arange(1000, 250000, 1.)  # in [Å]
            sed_flux = 3631e-23 * c / sed_wave**2  # in [erg s^-1 cm^-2 Å^-1]
        else:
            sed_file = os.path.join(pisco_path, "standards", standard_sed)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                sed_wave, sed_flux = np.loadtxt(sed_file).T
        f_sed = self.integrate_filter(sed_wave, sed_flux)

        return f_sed

    def _get_standard_mag(self, mag_sys_file):
        """Obtains the tabulated magnitude of a standard star.

        Parameters
        ----------
        mag_sys_file: str
            File with the magnitude system information.

        Returns
        -------
        sed_mag: float
            Tabulated standard magnitude.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            filt_names, mags = np.loadtxt(mag_sys_file, dtype=str).T
        err_message = f"{self.name} not in {mag_sys_file}"
        assert self.name in filt_names, err_message

        ind = list(filt_names).index(self.name)
        sed_mag = eval(mags[ind])

        return sed_mag

    def calculate_zp(self, mag_sys):
        """Calculates the zero point in the ``AB``, ``Vega`` or ``BD17``
        magnitude systems.

        Parameters
        ----------
        mag_sys : str
            Magnitude system. For example: ``AB``, ``BD17`` or ``Vega``.

        Returns
        -------
        zp : float
            Zero-point in the given natural magnitude system.
        """
        pisco_path = piscola.__path__[0]
        mag_sys_file_path = os.path.join(pisco_path, "mag_sys", "magnitude_systems.txt")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mag_sys_names, mag_sys_files = np.loadtxt(mag_sys_file_path, dtype=str).T
        err_message = f"mag. system {mag_sys} not found in {mag_sys_file_path}"
        assert mag_sys in mag_sys_names, err_message

        ind = list(mag_sys_names).index(mag_sys)
        mag_sys_file = os.path.join(pisco_path, "mag_sys", mag_sys_files[ind])

        f_sed = self._get_standard_flux(mag_sys_file)
        m_sed = self._get_standard_mag(mag_sys_file)
        zp = 2.5 * np.log10(f_sed) + m_sed

        return zp
    
    def calculate_extinction(self, ra, dec, scaling=0.86, reddening_law="fitzpatrick99", r_v=3.1, ebv=None):
        """Calculates the extinction for a given filter, given a right ascension and declination or :math:`E(B-V)`.

        Parameters
        ----------
        ra : float
            Right ascension in degrees.
        dec : float
            Declination in degrees.
        scaling: float, default ``0.86``
            Calibration of the Milky Way dust maps. Either ``0.86``
            for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
            dust map of Schlegel, Fikbeiner & Davis (1998).
        reddening_law: str, default ``fitzpatrick99``
            Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
            ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
            :math:`R_V = 3.1`.)
        r_v : float, default ``3.1``
            Total-to-selective extinction ratio (:math:`R_V`)
        ebv : float, default ``None``
            Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

        Returns
        -------
        A : float
            Extinction value in magnitudes.
        """

        A = extinction_filter(self.wavelength, self.transmission, ra, dec, 
                              scaling=scaling, reddening_law=reddening_law, 
                              r_v=r_v, ebv=ebv)
        
        return A


class MultiFilters(object):
    """Class representing multiple filters."""

    def __init__(self, bands, mag_systems):
        """
        Parameters
        ----------
        bands: list-like
            Bands to include. If ``None``, only include
            Bessell filters.
        mag_systems: list-like
            Magnitude systems. E.g. AB, Vega.
        """
        if bands is None:
            bands = []
            mag_systems = []
        self.bands = list(bands).copy()
        self.mag_systems = list(mag_systems).copy()

        for band, mag_sys in zip(bands, mag_systems):
            single_filt = SingleFilter(band, mag_sys)
            setattr(self, band, single_filt)

        # add Bessell filters
        filters = "UBVRI"
        for filt in filters:
            band = f"Bessell_{filt}"
            self._add_filter(band, 'BD17')

    def __repr__(self):
        return str(self.bands)

    def __getitem__(self, item):
        return getattr(self, item)

    def _add_filter(self, band, mag_sys):
        """Adds a band object with a single-filter class.

        Parameters
        ----------
        band: str
            Name of the band.
        """
        if band not in self.bands:
            single_filt = SingleFilter(band, mag_sys)
            setattr(self, band, single_filt)
            self.bands.append(band)
            self.mag_systems.append(mag_sys)

    def remove_filter(self, band):
        """Removes a band object.

        Parameters
        ----------
        band: str
            Name of the band.
        """
        err_message = f"Filter not found: {self.bands}"
        assert band in self.bands, err_message

        delattr(self, band)
        self.bands.remove(band)

    def calc_eff_wave(self, bands=None, sed_wave=None, sed_flux=None):
        """Calculates the effective wavelength of multiple bands.

        The effective wavelength depends on the shape
        of the given SED.

        Parameters
        ----------
        bands: list-like, default ``None``
            Bands for calculating their effective wavelengths. If ``None``,
            use of the available bands.
        sed_wave : array-like, default ``None``
            SED's wavelength range. If ``None``, a flat SED is used.
        sed_flux : array-like, default ``None``
            SED's flux density distribution. If ``None``, a flat SED is used.
        """
        if not bands:
            bands = self.bands

        for band in bands:
            self[band].calc_eff_wave(sed_wave, sed_flux)

    def calc_pivot(self, z=0.0):
        """Calculates the observed band closest to restframe
        :math:`B` band (:math:`4500 Å`).

        Parameters
        ----------
        z : float, default ``0.0``
            Redshift.
        """
        B_eff_wave = 4500.0
        eff_waves = np.array([self[band].eff_wave / (1 + z) for band in self.bands])
        idx = (np.abs(B_eff_wave - eff_waves)).argmin()
        self.pivot_band = self.bands[idx]

    def plot_filters(self, bands=None):
        """Plot the filters' transmission functions.

        Parameters
        ----------
        bands : list, default ``None``
            Bands to plot. If ``None``, plots all the available bands.
        """
        if not bands:
            bands = self.bands

        fig, ax = plt.subplots(figsize=(8, 6))
        for band in bands:
            norm = self[band]["transmission"].max()
            ax.plot(self[band]["wavelength"], self[band]["transmission"] / norm, label=band)

        ax.set_xlabel(r"Wavelength ($\AA$)", fontsize=18, family="serif")
        ax.set_ylabel("Normalised Transmission Function", fontsize=18, family="serif")
        ax.xaxis.set_tick_params(labelsize=16)
        ax.yaxis.set_tick_params(labelsize=16)
        ax.minorticks_on()
        ax.tick_params(
            which="major", length=6, width=1, direction="in", top=True, right=True
        )
        ax.tick_params(
            which="minor", length=3, width=1, direction="in", top=True, right=True
        )
        plt.legend(loc="upper right", bbox_to_anchor=(1.3, 1.0), fontsize=14)
        # plt.tight_layout()
        # plt.savefig('filters.png')

        plt.show()


def trim_filters(response):
    """Trim the leading and trailing zeros from a 1-D array or sequence, leaving
    one zero on each side. This is a modified version of :func:`numpy.trim_zeros`.

    Parameters
    ----------
    response : 1-D array or sequence
        Input array.

    Returns
    -------
    first : int
        Index of the last leading zero.
    last : int
        Index of the first trailing zero.
    """

    first = 0
    for i in response:
        if i != 0.0:
            if first == 0:
                first += 1  # to avoid filters with non-zero edges
            break
        else:
            first = first + 1

    last = len(response)
    for i in response[::-1]:
        if i != 0.0:
            if last == len(response):
                last -= 1  # to avoid filters with non-zero edges
            break
        else:
            last = last - 1

    first -= 1
    last += 1

    return first, last



import unittest
from piscola.filters_class import MultiFilters
import matplotlib.pyplot as plt


class TestPiscola(unittest.TestCase):
    def test_plot(self):
        filters = MultiFilters(["ps1_g"], ["AB"])
        plt.ion()
        filters.plot_filters()

    def test_remove_filter(self):
        filters = MultiFilters(["ps1_g"], ["AB"])
        filters.remove_filter("ps1_g")

    def test_pivot_wavelength(self):
        filters = MultiFilters(["ps1_g"], ["AB"])
        filters.calc_pivot()


if __name__ == "__main__":
    unittest.main()



import jax
import jaxopt
import numpy as np
import jax.numpy as jnp
from tinygp import GaussianProcess, kernels, transforms

jax.config.update("jax_enable_x64", True)

def prepare_gp_inputs(times, wavelengths, fluxes, flux_errors, fit_type, wave_log):
    """Prepares the inputs for the Gaussian Process model fitting.
    
    Parameters
    ----------
    times: ndarray 
        Light-curve epochs.
    wavelengths: ndarray 
        Light-curve effective wavelength.
    fluxes: ndarray 
        Light-curve fluxes.
    flux_errors: ndarray 
        Light-curve flux errors.
    fit_type: str
        Transformation used for the light-curve fits: ``flux``, ``log``, ``arcsinh``.
    wave_log: bool
        Whether to use logarithmic (base 10) scale for the 
        wavelength axis.
        
    Returns
    -------
    X: ndarray
        X-axis array for the Gaussian Process model.
    y: ndarray
        Y-axis array for the Gaussian Process model.
    yerr: ndarray
        Y-axis errors for the Gaussian Process model.
    y_norm: float
        Normalisation used for the fluxes and errors. The maximum
        of the fluxes is used.
    """
    valid_fit_types = ["flux", "log", "arcsinh"]
    err_message = f"Not a valid fit type ({valid_fit_types}): {fit_type}"
    assert fit_type in valid_fit_types, err_message

    X = (times, wavelengths)
    if wave_log is True:
        X = (times, jnp.log10(wavelengths) * 10)

    # normalise fluxes - values have to be ideally above zero
    y_norm = np.copy(fluxes.max()) 
    if fit_type == "flux":
        y = (fluxes / y_norm).copy() 
        yerr = (flux_errors / y_norm).copy()
    elif fit_type == "log":
        mask = fluxes > 0.0
        #y_norm = np.copy(fluxes[mask].mean())
        if len(fluxes) == len(X[0]):
            # when predicting, the arrays do not match in length
            X = (X[0][mask], X[1][mask])
        y = np.log10(fluxes[mask] / y_norm)
        yerr = np.abs(flux_errors / (fluxes * np.log(10)))[mask]
    elif fit_type == "arcsinh":
        if len(fluxes) == len(X[0]):
            # when predicting, the arrays do not match in length
            X = (X[0], X[1])
        y = np.arcsinh(fluxes / y_norm) 
        yerr = (flux_errors / y_norm) / np.sqrt((fluxes / y_norm) ** 2 + 1)

    return X, y, yerr, y_norm

def fit_gp_model(times, wavelengths, fluxes, flux_errors, k1='Matern52', fit_mean=False, 
                 fit_type='flux', wave_log=True, time_scale=None, wave_scale=None, add_noise=True):
    """Fits a Gaussian Process model to a SN multi-colour light curve.
    
    All input arrays MUST have the same length.

    Parameters
    ----------
    times: ndarray 
        Light-curve epochs.
    wavelengths: ndarray 
        Light-curve effective wavelength.
    fluxes: ndarray 
        Light-curve fluxes.
    flux_errors: ndarray 
        Light-curve flux errors.
    k1: str
        Kernel to be used for the time axis. Either ``Matern52``,
        ``Matern32`` or ``ExpSquared``.
    fit_type: str
        Transformation used for the light-curve fits: ``flux``, ``log``, ``arcsinh``.
    wave_log: bool, default ``True``.
        Whether to use logarithmic (base 10) scale for the 
        wavelength axis.
    time_scale: float, default ``None``
        If given, the time scale is fixed using this value, in units of days.
    wave_scale: float, default ``None``
        If given, the wavelength scale is fixed using this value, in units of angstroms.
        Note that if 'wave_log=True', the logarithm base 10 of this value is used.
    add_noise: bool, default ``True``
        Whether to add a white-noise component to the GP model. This "inflates" the errors.
        
    Returns
    -------
    gp_model: ~tinygp.gp.GaussianProcess
        Gaussian Process multi-colour light-curve model.
    """
    assert k1 in ['Matern52', 'Matern32', 'ExpSquared'], "Not a valid kernel"
    def build_gp(params):
        """Creates a Gaussian Process model.
        """
        nonlocal time_scale, wave_scale, add_noise  # import from main function
        if time_scale is None:
            log_time_scale = params["log_scale"][0]
        else:
            log_time_scale = np.log(time_scale)
        if wave_scale is None:
            log_wave_scale = params["log_scale"][-1]
        else:
            log_wave_scale = np.log(wave_scale)
        if add_noise is True:
            noise = jnp.exp(2 * params["log_noise"])
        else:
            noise = 0.0

        # select time-axis kernel
        if k1 == 'Matern52':
            kernel1 = transforms.Subspace(0, kernels.Matern52(scale=jnp.exp(log_time_scale)))
        elif k1 == 'Matern32':
            kernel1 = transforms.Subspace(0, kernels.Matern32(scale=jnp.exp(log_time_scale)))
        else:
            kernel1 = transforms.Subspace(0, kernels.ExpSquared(scale=jnp.exp(log_time_scale)))
        # wavelength-axis kernel
        kernel2 = transforms.Subspace(1, kernels.ExpSquared(scale=jnp.exp(log_wave_scale)))
        
        kernel = jnp.exp(params["log_amp"]) * kernel1 * kernel2
        diag = yerr ** 2 + noise
        
        if fit_mean is True:
            mean = jnp.exp(params["log_mean"])
        elif fit_type == "log":
            mean = -jnp.exp(params["log_mean"])
        else:
            mean = None

        return GaussianProcess(kernel, X, diag=diag, mean=mean)

    @jax.jit
    def loss(params):
        """Loss function for the Gaussian Process hyper-parameters optimisation.
        """
        return -build_gp(params).condition(y).log_probability
    
    X, y, yerr, _ = prepare_gp_inputs(times, wavelengths, fluxes, 
                                      flux_errors, fit_type=fit_type, 
                                      wave_log=wave_log)

    # GP hyper-parameters
    scales = np.array([30, 2000]) # units: days, angstroms
    if wave_log is True:
        # the approx range from opt to NIR is 1 in log space
        scales = np.array([30, 0.5]) # units: days, log10(angstroms)
    if time_scale is not None:
        scales = np.delete(scales, 0)
    if wave_scale is not None:
        scales = np.delete(scales, -1)
    
    # create parameters object
    params = {
        "log_amp": jnp.log(y.var()),
        "log_noise": jnp.log(np.mean(yerr)),
    }
    if len(scales) != 0:
        params.update({"log_scale": jnp.log(scales)})
    if fit_mean is True:
        params.update({"log_mean": jnp.log(np.average(y, weights=1/yerr**2))})
    elif fit_type == "log":
        # absolute value to avoid negative in log
        params.update({"log_mean": jnp.log(np.abs(y.min()))})

    # Train the GP model
    solver = jaxopt.ScipyMinimize(fun=loss)
    soln = solver.run(params)
    gp_model = build_gp(soln.params)
    # store initial and optimised parameters
    gp_model.__dict__["init_params"] = params
    gp_model.__dict__["params"] = soln.params
    
    return gp_model

def fit_single_lightcurve(times, fluxes, flux_errors, k1='Matern52'):
    """Fits a Gaussian Process model to a single SN light curve.
    
    All input arrays MUST have the same length.

    Parameters
    ----------
    times: ndarray 
        Light-curve epochs.
    fluxes: ndarray 
        Light-curve fluxes.
    flux_errors: ndarray 
        Light-curve flux errors.
    k1: str
        Kernel to be used for the time axis. Either ``Matern52``,
        ``Matern32`` or ``ExpSquared``.
        
    Returns
    -------
    gp_model: ~tinygp.gp.GaussianProcess
        Gaussian Process light-curve model.
    """
    assert k1 in ['Matern52', 'Matern32', 'ExpSquared'], "Not a valid kernel"
    def build_gp(params):
        """Creates a Gaussian Process model.
        """
        # select time-axis kernel
        if k1 == 'Matern52':
            kernel1 = kernels.Matern52(scale=jnp.exp(params["log_scale"]))
        elif k1 == 'Matern32':
            kernel1 = kernels.Matern32(scale=jnp.exp(params["log_scale"]))
        else:
            kernel1 = kernels.ExpSquared(scale=jnp.exp(params["log_scale"]))
        kernel = jnp.exp(params["log_amp"]) * kernel1
        diag = yerr ** 2 

        return GaussianProcess(kernel, x, diag=diag)

    @jax.jit
    def loss(params):
        """Loss function for the Gaussian Process hyper-parameters optimisation.
        """
        return -build_gp(params).condition(y).log_probability
    
    x = times.copy()
    y = fluxes.copy() / fluxes.max()
    yerr = flux_errors.copy() / fluxes.max()
    
    # GP hyper-parameters
    time_scale = 10  # days
    params = {
        "log_amp": jnp.log(y.var()),
        "log_scale": jnp.log(time_scale),
    }
    # Train the GP model
    solver = jaxopt.ScipyMinimize(fun=loss)
    soln = solver.run(params)
    gp_model = build_gp(soln.params)
    
    return gp_model


# -*- coding: utf-8 -*-
# This is the skeleton of PISCOLA, the main file
import os
import bz2
import sys
import math
import statistics
import numpy as np
import pandas as pd
from peakutils import peak

if sys.version_info.minor < 8:
    import pickle5 as pickle
else:
    # pickle is a built-in package starting from python v3.8
    import pickle

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec

from .utils import change_zp, flux_to_mag, calculate_robust_stats
from .filters_class import MultiFilters
from .lightcurves_class import Lightcurves
from .extinction_correction import extinction_at_wavelength
from .gaussian_process import prepare_gp_inputs, fit_gp_model

import jax
jax.config.update("jax_enable_x64", True)

def call_sn(lc_file):
    """Loads a supernova from a file and initialises
    the :func:`sn` object. The object is initialised
    with all the necessary information like filters,
    fluxes, etc.

    Parameters
    ----------
    lc_file : str
        File with the SN info and light curves.

    Returns
    -------
    sn_obj : obj
        New :func:`sn` object.
    """
    err_message = f"File {lc_file} not found."
    assert os.path.isfile(lc_file), err_message

    converters = {"name": str, "z": float, "ra": float, "dec": float}
    name, z, ra, dec = pd.read_csv(
        lc_file, nrows=1, sep='\\s+', converters=converters
    ).values[0]

    sn_obj = Supernova(name, z, ra, dec, lc_file)

    return sn_obj


def load_sn(piscola_file):
    """Loads a :func:`sn` oject that was previously saved as a pickle file.

    Parameters
    ----------
    piscola_file : str
        File with the SN object saved with PISCOLA.

    Returns
    -------
    pickle.load(file) : obj
        :func:`sn` object previously saved as a pickle file.
    """
    err_message = f"File {piscola_file} not found."
    assert os.path.isfile(piscola_file), err_message

    with bz2.BZ2File(piscola_file, "rb") as sn_file:
        sn_obj = pickle.load(sn_file)

    return sn_obj


class Supernova(object):
    """Class representing a supernova object."""

    def __init__(
        self, name, z=0.0, ra=None, dec=None, lc_file=None
    ):
        """
        Parameters
        ----------
        name: str
            Name of the supernova.
        z: float, default ``0.0``
            Redshift of the supernova.
        ra: float, default ``None``
            Right ascension of the supernova.
        dec: float, default ``None``
            Declination of the supernova.
        lc_file: str, default ``None``
            File with the supernova light-curve data.
        """
        self.name = name
        self.z = z
        self.ra = ra
        self.dec = dec

        if not self.ra or not self.dec:
            print("Warning, RA and/or DEC not specified.")

        # add light curves and filters
        if lc_file:
            lcs_df = pd.read_csv(lc_file, sep='\\s+', skiprows=2)
            # saves original light curves untouched
            self.init_lcs = Lightcurves(lcs_df)
            # these are the light curves that will be used
            self.lcs = Lightcurves(lcs_df)
            mag_systems = [self.lcs[band].mag_sys for band in self.lcs.bands]
            self.filters = MultiFilters(self.lcs.bands, mag_systems)

            # order bands by effective wavelength
            eff_waves = [self.filters[band]["eff_wave"] for band in self.filters.bands]
            sorted_idx = sorted(range(len(eff_waves)), key=lambda k: eff_waves[k])
            sorted_bands = [self.filters.bands[x] for x in sorted_idx]
            lc_bands = [band for band in sorted_bands if band in lcs_df.band.values]
            self.bands = self.lcs.bands = lc_bands
            self.filters.bands = sorted_bands
            self._normalize_lcs()

    def __repr__(self):
        rep = (
            f"name: {self.name}, z: {self.z:.5}, "
            f"ra: {self.ra:.5}, dec: {self.dec:.5}\n"
        )
        return rep

    def __getitem__(self, item):
        return getattr(self, item)
    
    def add_filter(self, band, mag_sys):
        """Adds a new filter.

        Parameters
        ----------
        band : str
            Name of the filter
        mag_sys : str
            Magnitude system.
        """
        self.filters._add_filter(band, mag_sys)
        # order bands by effective wavelength
        eff_waves = [self.filters[band]["eff_wave"] for band in self.filters.bands]
        sorted_idx = sorted(range(len(eff_waves)), key=lambda k: eff_waves[k])
        sorted_bands = [self.filters.bands[x] for x in sorted_idx]
        self.filters.bands = sorted_bands

    def save(self, path=None):
        """Saves a SN object into a pickle file

        Parameters
        ----------
        path: str, default ``None``
            Path where to save the SN object. If ``None``,
            use current directory. The file will be saved with
            the name ``<SN name>.pisco``.
        """
        if path is None:
            path = ""

        outfile = os.path.join(path, f"{self.name}.pisco")
        with bz2.BZ2File(outfile, "wb") as pfile:
            protocol = 4
            pickle.dump(self, pfile, protocol)

    def _normalize_lcs(self):
        """Normalizes the fluxes and zero-points (ZPs).

        Fluxes are converted to physical units by calculating the ZPs according to the
        magnitude system, for example: **AB**, **BD17** or **Vega**.
        """
        for band in self.bands:
            mag_sys = self.lcs[band].mag_sys
            current_zp = self.lcs[band].zp

            new_zp = self.filters[band].calculate_zp(mag_sys)

            self.lcs[band].fluxes = change_zp(self.lcs[band].fluxes, current_zp, new_zp)
            self.lcs[band].flux_errors = change_zp(
                self.lcs[band].flux_errors, current_zp, new_zp
            )
            self.lcs[band].zp = new_zp

    ####################
    # Light-curve fits #
    ####################
    def _stack_lcs(self, bands=None):
        """Stacks of light-curve properties

        Times, wavelengths, fluxes, magnitudes and errors are
        stacked for 2D fitting.

        Parameters
        ----------
        bands: list-like
            Bands used for the stacking.
        """
        if bands is None:
            stacking_bands = self.bands
        else:
            stacking_bands = bands

        # convert the data from multiple light curves into single arrays
        times = np.hstack([self.lcs[band].times for band in stacking_bands])
        wavelengths = np.hstack(
            [
                [self.filters[band].eff_wave] * len(self.lcs[band].times)
                for band in stacking_bands
            ]
        )
        fluxes = np.hstack([self.lcs[band].fluxes for band in stacking_bands])
        flux_errors = np.hstack([self.lcs[band].flux_errors for band in stacking_bands])
        magnitudes = np.hstack([self.lcs[band].magnitudes for band in self.bands])
        mag_errors = np.hstack([self.lcs[band].mag_errors for band in stacking_bands])

        # store in hidden variables for GP fitting
        self._stacked_times = times
        self._stacked_wavelengths = wavelengths
        self._stacked_fluxes = fluxes
        self._stacked_flux_errors = flux_errors
        self._stacked_magnitudes = magnitudes
        self._stacked_mag_errors = mag_errors

    def gp_predict(self, times_pred, wavelengths_pred, return_cov=True):
        """Returns the Gaussian Process prediction of the light-curve fits.

        Parameters
        ----------
        times_pred : ndarray
            Time-axis grid.
        wavelengths_pred : ndarray
            Wavelength-axis grid.
        return_cov : bool, optional
            Whether to return the covariance or variance.

        Returns
        -------
        mu: ndarray
            Mean of the Gaussian Process.
        cov: ndarray
            Either the variance or covariance of the Gaussian Process.
        """
        X_test, y, _, y_norm = prepare_gp_inputs(times_pred, wavelengths_pred, 
                                                self._stacked_fluxes, self._stacked_flux_errors, 
                                                fit_type=self.fit_type,
                                                wave_log=self.wave_log)

        # Function to propagate the covariance matrix through nonlinear functions
        def transform_covariance(cov_matrix, x, fit_type):
            # Compute the Jacobian at point x
            if fit_type == "log":
                jac = np.diag((10 ** x) * np.log(10))
            elif fit_type == "arcsinh":    
                jac = np.diag(1 / np.sqrt(x**2 + 1))
            # Transform the covariance matrix using the Jacobian
            transformed_cov_matrix = jac @ cov_matrix @ jac.T
            return transformed_cov_matrix

        # GP prediction
        if return_cov is True:
            return_var = False
        else:
            return_var = True

        # if return_cov and return_var are both True, the covariance flag is ignored
        if self.fit_type == "flux":
            mu, cov = self.gp_model.predict(y, X_test=X_test, return_var=return_var, return_cov=True)
        elif self.fit_type == "log":
            log_mu, log_cov = self.gp_model.predict(y, X_test=X_test, return_var=return_var, return_cov=True)
            mu = 10 ** log_mu
            cov = transform_covariance(log_cov, log_mu, self.fit_type)
        elif self.fit_type == "arcsinh":
            arcsinh_mu, arcsinh_cov = self.gp_model.predict(y, X_test=X_test, return_var=return_var, return_cov=True)
            mu = np.sinh(arcsinh_mu)
            cov = transform_covariance(arcsinh_cov, arcsinh_mu, self.fit_type)

        # renormalise outputs and convert jax-Array to numpy-array
        mu = np.array(mu) * y_norm
        cov = np.array(cov) * y_norm ** 2

        return mu, cov
        
    def fit(self, bands=None, k1='Matern52', fit_type='flux', wave_log=True, time_scale=None, wave_scale=None, add_noise=True):
        """Fits the observed multi-colour light-curve data with Gaussian Process.

        The time of rest-frame B-band peak luminosity is estimated by finding where the derivative is equal to zero.

        Parameters
        ----------
        bands : list-like, default ``None``
            Bands used for fitting light curves. By default, use all the available bands.
        k1: str
            Kernel to be used for the time axis. Either ``Matern52``,
            ``Matern32`` or ``ExpSquared``.
        fit_type: str, default ``flux``.
            Transformation used for the light-curve fits: ``flux``, ``log``, ``arcsinh``.
        wave_log: bool, default ``True``.
            Whether to use logarithmic (base 10) scale for the 
            wavelength axis.
        time_scale: float, default ``None``
            If given, the time scale is fixed using this value, in units of days.
        wave_scale: float, default ``None``
            If given, the wavelength scale is fixed using this value, in units of angstroms.
            Note that if 'wave_log=True', the logarithm base 10 of this value is used.
        add_noise: bool, default ``True``
            Whether to add a white-noise component to the GP model. This "inflates" the errors.
        """
        ##########
        # GP fit #
        ##########
        self._stack_lcs(bands)
        gp_model = fit_gp_model(self._stacked_times, self._stacked_wavelengths, 
                                self._stacked_fluxes, self._stacked_flux_errors, k1=k1, 
                                fit_type=fit_type, wave_log=wave_log, time_scale=time_scale, 
                                wave_scale=wave_scale, add_noise=add_noise)
        # store GP model and input parameters
        self.gp_model = gp_model
        self.k1 = k1
        self.fit_type = fit_type
        self.wave_log = wave_log
        self.time_scale = time_scale
        self.wave_scale = wave_scale
        self.add_noise = add_noise

        ############################
        # Light-curves predictions #
        ############################
        dt = 0.1 * (1 + self.z)  # 0.1 days in rest-frame, moved to observer frame
        times_pred = np.arange(self._stacked_times.min() - 5, 
                               self._stacked_times.max() + 5,
                               dt
                               )
        self.times_pred = times_pred

        fitting_bands = bands
        if bands is None:
            fitting_bands = self.bands
            
        fits_df_list = []
        for band in fitting_bands:
            eff_wave = self.filters[band]["eff_wave"]
            wavelengths_pred = np.zeros_like(times_pred) + eff_wave
            mu, var = self.gp_predict(times_pred, wavelengths_pred, return_cov=False)
            std = np.sqrt(var)

            # store light-curve fits
            fit_df = pd.DataFrame({"time": times_pred, "flux": mu, "flux_err": std})
            fit_df["zp"] = self.lcs[band].zp
            fit_df["band"] = band
            fit_df["mag_sys"] = self.lcs[band].mag_sys
            fits_df_list.append(fit_df)

        self.lc_fits = Lightcurves(pd.concat(fits_df_list))

        # get rest-frame light curve and light-curve parameters
        self._get_rest_lightcurves()

    def _get_rest_lightcurves(self, bands=None):
        """Obtains the corrected rest-frame light curves from the fits.

        Returns
        -------
        bands : list-like, default ``None``
            Bands for which to extract the light curves. All bands are used by default.
        """
        # always add Bessell_B band
        if bands is None:
            bands = self.lcs.bands
        if 'Bessell_B' not in bands:
            bands = ['Bessell_B'] + bands
        self.lc_parameters = {'tmax':None, 'tmax_err':None, 
                                'mmax':None, 'mmax_err':None, 
                                'dm15':None, 'dm15_err':None, 
                                'colour':None, 'colour_err':None, 
                                }

        rest_df_list = []
        rest_times = self.times_pred / (1 + self.z)

        for band in bands:
            rest_eff_wave = self.filters[band].eff_wave * (1 + self.z)
            wavelengths_pred = np.zeros_like(self.times_pred) + rest_eff_wave            
            mu, cov = self.gp_predict(self.times_pred, wavelengths_pred, return_cov=True)

            # correct for MW dust extinction and redshift 
            A = extinction_at_wavelength(rest_eff_wave, self.ra, self.dec)
            correction = 10**(0.4*A) * (1 + self.z)
            mu *= correction
            cov *= correction ** 2
            std = np.sqrt(np.diag(cov))

            self.lc_parameters.update({band:{}})
            try:
                # calculate light-curve parameters using restframe epochs
                _, tmax, tmax_err, fmax, fmax_err, df15, df15_err = self._calculate_lc_params(rest_times, mu, cov)
                # tmax - this actually needs to be corrected for redshift
                tmax *= (1 + self.z)
                tmax_err *= (1 + self.z)
                self.lc_parameters[band]['tmax'] = np.round(tmax, 3)
                self.lc_parameters[band]['tmax_err'] = np.round(tmax_err, 3)
                # mmax
                zp = self.filters[band].calculate_zp(self.filters[band].mag_sys)
                mmax, mmax_err = flux_to_mag(fmax, fmax_err, zp)
                self.lc_parameters[band]['mmax'] = np.round(mmax, 3)
                self.lc_parameters[band]['mmax_err'] = np.round(mmax_err, 3)
                # dm15 
                dm15, dm15_err = flux_to_mag(df15, df15_err, 0.0)
                self.lc_parameters[band]['dm15'] = np.round(dm15, 3)
                self.lc_parameters[band]['dm15_err'] = np.round(dm15_err, 3)
            except:
                self.lc_parameters[band]['tmax'] = self.lc_parameters[band]['tmax_err'] = np.nan
                self.lc_parameters[band]['mmax'] = self.lc_parameters[band]['mmax_err'] = np.nan
                self.lc_parameters[band]['dm15'] = self.lc_parameters[band]['dm15_err'] = np.nan

            if band == 'Bessell_B':
                # store B-band parameters in an "upper level"
                for key, val in self.lc_parameters[band].items():
                    self.lc_parameters[key] = val
                self.tmax = self.lc_parameters['tmax']
                self.tmax_err = self.lc_parameters['tmax_err']
                ######################
                # Colour Calculation #
                ######################
                # GP prediction
                band2 = 'Bessell_V'  # for (B-V)
                zp2 = self.filters[band2].calculate_zp(self.filters[band2].mag_sys)
                rest_eff_wave2 = self.filters[band2].eff_wave * (1 + self.z)
                wavelengths_pred = np.array([rest_eff_wave, rest_eff_wave2])
                times_pred_V = np.zeros_like(wavelengths_pred) + self.tmax
                mu2, cov2 = self.gp_predict(times_pred_V, wavelengths_pred, return_cov=True)

                # propagate MW dust extinction correction and redshift
                A2 = extinction_at_wavelength(rest_eff_wave2, self.ra, self.dec)
                corr_array = np.array([10 ** (0.4 * A), 10 ** (0.4 * A2)]) * (1 + self.z)
                mu2 *= corr_array
                cov2 *= np.outer(corr_array, corr_array)
                std2 = np.sqrt(np.diag(cov2))

                # calculate colour
                flux_ratio = mu2[0] / mu2[1]
                flux_ratio_error = np.abs(flux_ratio) * np.sqrt((std2[0] / mu2[0]) ** 2 + (std2[1] / mu2[1]) ** 2 - 2 * (cov2[0][1] / (mu2[0] * mu2[1])))
                colour, colour_err = flux_to_mag(flux_ratio, flux_ratio_error, zp=(zp - zp2))
                self.lc_parameters['colour'] = np.round(colour, 3)
                self.lc_parameters['colour_err'] = np.round(colour_err, 3)

            # store rest-frame light-curves
            rest_df = pd.DataFrame({"time": self.times_pred, "flux": mu, "flux_err": std})
            rest_df["zp"] = zp
            rest_df["band"] = band
            rest_df["mag_sys"] = self.filters[band].mag_sys
            rest_df_list.append(rest_df)

        self.rest_lcs = Lightcurves(pd.concat(rest_df_list))

    def _calculate_lc_params(self, rest_times, mu, cov):
        """Calculate rest-frame light-curve parameters."""
        ###############################
        # Estimate peak time and flux #
        ###############################
        # in the NIR, the SNe Ia peaks should be at least separated by around 10 days
        dt = (rest_times.max() - rest_times.min()) / 10
        min_dist = np.argmin(np.abs(rest_times - (rest_times.min() + dt)))
        peak_ids = peak.indexes(mu, thres=0.3, min_dist=min_dist)
        if len(peak_ids) == 0:
            # if no peak is found, just use the maximum
            max_id = np.argmax(mu)
        else:
            max_id = peak_ids[0]
        # get error on epoch at maximum by monte-carlo sampling
        lcs_sample = np.random.multivariate_normal(mu, cov, size=6000)
        tmax_list = []
        for lc in lcs_sample:
            peak_ids = peak.indexes(lc, thres=0.3, min_dist=min_dist)
            if len(peak_ids) == 0:
                peak_id = np.argmax(lc)
            else:
                peak_id = peak_ids[0]
            tmax_list.append(rest_times[peak_id])
        # store peak time
        tmax = rest_times[max_id]
        _, tmax_err, _ = calculate_robust_stats(tmax_list, sigma=5)
        # store flux at maximum
        fmax = mu[max_id]
        std = np.sqrt(np.diag(cov))
        fmax_err = std[max_id]

        ####################
        # Estimate Stretch #
        ####################
        t15 = rest_times[max_id] + 15
        ind15 = np.argmin(np.abs(rest_times - t15))
        f15 = mu[ind15]
        f15_err = std[ind15]
        df15 = f15 / fmax
        # error propagation
        f15_cov = cov[max_id][ind15]  # covariance, in flux, at peak and 15 days
        df15_err = np.abs(df15) * np.sqrt((f15_err / f15) ** 2 + (fmax_err / fmax) ** 2 - 2 * (f15_cov / (f15 * fmax)))

        return max_id, tmax, tmax_err, fmax, fmax_err, df15, df15_err
    
    def _calculate_lc_params2(self, rest_times, mu, cov):
        """Calculate light-curve parameters with the time in restframe."""
        # THIS PROBABLY NEEDS TO RETURN JUST THE PEAK INDECES
        # Calculate values by monte-carlo sampling
        lcs_sample = np.random.multivariate_normal(mu, cov, size=6000)
        tmax_list, fmax_list, df15_list = [], [], []
        maxid_list = []
        # in the NIR, the SNe Ia peaks should be at least separated by around 10 days
        dt = (rest_times.max() - rest_times.min()) / 10
        min_dist = np.argmin(np.abs(rest_times - (rest_times.min() + dt)))
        for fluxes in lcs_sample:
            peak_ids = peak.indexes(fluxes, thres=0.3, min_dist=min_dist)
            if len(peak_ids) == 0:
                # if no peak is found, just use the maximum
                max_id = np.argmax(fluxes)
            else:
                max_id = peak_ids[0]
            maxid_list.append(max_id)
            tmax_list.append(rest_times[max_id])
            fmax_list.append(fluxes[max_id])
            f15_id = np.argmin(np.abs(rest_times - (rest_times[max_id] + 15)))
            df15_list.append(fluxes[f15_id] / fluxes[max_id])
        # lists to arrays
        tmax_array = np.array(tmax_list)
        fmax_array = np.array(fmax_list)
        df15_array = np.array(df15_list)
        # obtain values
        max_id = statistics.mode(maxid_list)
        tmax, tmax_err, robust_mask = calculate_robust_stats(tmax_array, sigma=5)
        fmax, fmax_err = np.nanmean(fmax_array[robust_mask]), np.nanstd(fmax_array[robust_mask])
        df15, df15_err = np.nanmean(df15_array[robust_mask]), np.nanstd(df15_array[robust_mask])

        return max_id, tmax, tmax_err, fmax, fmax_err, df15, df15_err

    ################### 
    # Extra Functions #
    ###################
    # this are not essential but help assessing the fits
    # and output results, among other things
    def get_obs_params(self, bands=None):
        """Obtains the time and magnitude of maximum for the observed light curves.

        No corrections are applied at all.

        Returns
        -------
        bands : list-like, default ``None``
            Bands for which to extract the parameters.
        """
        if bands is None:
            bands = self.bands

        phases = (self.times_pred - self.tmax) / (1 + self.z)
        self.obs_parameters = {band:{} for band in bands}
        for band in bands:
            obs_eff_wave = self.filters[band].eff_wave
            wavelengths_pred = np.zeros_like(self.times_pred) + obs_eff_wave
            mask = (-15 <= phases) & (phases <= 10)  # mask for quicker calculation
            mu, cov = self.gp_predict(self.times_pred[mask], wavelengths_pred[mask], return_cov=True)

            # calculate light-curve parameters
            _, tmax, tmax_err, fmax, fmax_err, _, _ = self._calculate_lc_params(self.times_pred[mask], mu, cov)
            # tmax - taken from the initial light-curve fit
            self.obs_parameters[band]['tmax'] = np.round(tmax, 3)
            self.obs_parameters[band]['tmax_err'] = np.round(tmax_err, 3)
            # mmax
            zp = self.filters[band].calculate_zp(self.filters[band].mag_sys)
            mmax, mmax_err = flux_to_mag(fmax, fmax_err, zp)
            self.obs_parameters[band]['mmax'] = np.round(mmax, 3)
            self.obs_parameters[band]['mmax_err'] = np.round(mmax_err, 3)

    def plot_lcs(self, plot_mag=False, fig_name=None):
        """Plots the observed light-curves.

        Parameters
        ----------
        plot_mag : bool, default ``False``
            If ``True``, plots the bands in magnitude space.
        fig_name : str, default ``None``
            If  given, name of the output plot.
        """
        palette1 = [plt.get_cmap("Dark2")(i) for i in np.arange(8)]
        palette2 = [plt.get_cmap("Set1")(i) for i in np.arange(9)]
        palette3 = [plt.get_cmap("Pastel2")(i) for i in np.arange(8)]
        colours = palette1 + palette2 + palette3 + palette1 + palette2

        # shift in time for visualization purposes
        min_time = self.lcs[self.bands[1]].times.min()
        tmax_str = str(min_time.astype(int))
        zeros = "0" * len(tmax_str[2:])
        t_offset = int(tmax_str[:2] + zeros)

        ZP = 27.5  # global zero-point for visualization

        h = 3  # columns
        v = math.ceil(len(self.bands) / h)  # rows

        fig = plt.figure(figsize=(15, 5 * v))
        gs = gridspec.GridSpec(v, h)

        for i, band in enumerate(self.bands):
            j = math.ceil(i % h)
            k = i // h
            ax = plt.subplot(gs[k, j])

            x = self.lcs[band].times - t_offset
            if plot_mag is False:
                y_norm = change_zp(1.0, self.lcs[band].zp, ZP)
                y = self.lcs[band].fluxes * y_norm
                yerr = self.lcs[band].flux_errors * y_norm
            else:
                y = self.lcs[band].magnitudes
                yerr = self.lcs[band].mag_errors
                ax.invert_yaxis()

            colour = colours[i]
            data_par = dict(
                fmt="o",
                color=colour,
                capsize=3,
                capthick=2,
                ms=8,
                elinewidth=3,
                mec="k",
            )

            # light curves
            ax.errorbar(x, y, yerr, label=band, **data_par)

            ax.xaxis.set_tick_params(labelsize=15)
            ax.yaxis.set_tick_params(labelsize=15)
            ax.minorticks_on()
            ax.tick_params(
                which="major", length=6, width=1, direction="in", top=True, right=True
            )
            ax.tick_params(
                which="minor", length=3, width=1, direction="in", top=True, right=True
            )
            ax.legend(fontsize=16)

        fig.text(
            0.5,
            0.92,
            f"{self.name} (z = {self.z:.5})",
            ha="center",
            fontsize=20,
            family="serif",
        )
        fig.text(
            0.5,
            0.05,
            f"Time - {t_offset} [days]",
            ha="center",
            fontsize=18,
            family="serif",
        )
        if not plot_mag:
            fig.text(
                0.05,
                0.5,
                f"Flux (ZP = {ZP})",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )
        else:
            fig.text(
                0.05,
                0.5,
                r"Apparent Magnitude",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )

        if fig_name is not None:
            plt.savefig(fig_name)

        plt.show()

    def plot_fits(self, plot_mag=False, fig_name=None):
        """Plots the light-curves fits results.

        Plots the observed data for each band together with the gaussian process fits.
        The time of :math:`B`-band peak is shown as a vertical dashed line.

        Parameters
        ----------
        plot_mag : bool, default ``False``
            If ``True``, plots the bands in magnitude space.
        fig_name : str, default ``None``
            If  given, name of the output plot.
        """
        palette1 = [plt.get_cmap("Dark2")(i) for i in np.arange(8)]
        palette2 = [plt.get_cmap("Set1")(i) for i in np.arange(9)]
        palette3 = [plt.get_cmap("Pastel2")(i) for i in np.arange(8)]
        colours = palette1 + palette2 + palette3 + palette1 + palette2

        # shift in time for visualization purposes
        try:
            tmax_str = str(self.tmax.astype(int))
        except:
            tmax_str = str(self.tmax.astype(int))
        zeros = "0" * len(tmax_str[2:])
        t_offset = int(tmax_str[:2] + zeros)

        # data and fits
        data = self.lcs
        fits = self.lc_fits
        tmax = self.tmax

        ZP = 27.5  # global zero-point for visualization

        h = 3  # columns
        v = math.ceil(len(fits.bands) / h)  # rows

        fig = plt.figure(figsize=(15, 5 * v))
        gs = gridspec.GridSpec(v * 2, h, height_ratios=[3, 1] * v)

        for i, band in enumerate(fits.bands):
            j = math.ceil(i % h)
            k = i // h * 2
            ax = plt.subplot(gs[k, j])
            ax2 = plt.subplot(gs[k + 1, j])

            x = data[band].times - t_offset
            x_fit = fits[band].times - t_offset

            if plot_mag is False:
                y_norm = change_zp(1.0, data[band].zp, ZP)
                y = data[band].fluxes * y_norm
                yerr = data[band].flux_errors * y_norm
                y_fit = fits[band].fluxes * y_norm
                yerr_fit = fits[band].flux_errors * y_norm
            else:
                y = data[band].magnitudes
                yerr = data[band].mag_errors
                y_fit = fits[band].magnitudes
                yerr_fit = fits[band].mag_errors
                ax.invert_yaxis()
                ax2.invert_yaxis()

            colour = colours[i]
            data_par = dict(
                fmt="o",
                color=colour,
                capsize=3,
                capthick=2,
                ms=8,
                elinewidth=3,
                mec="k",
            )
            fit_par = dict(ls="-", lw=2, zorder=16, color=colour)
            fit_err_par = dict(alpha=0.3, color=colour)

            # light curves
            ax.errorbar(x, y, yerr, label=band, **data_par)
            ymin, ymax = ax.get_ylim()
            ax.plot(x_fit, y_fit, **fit_par)
            ax.fill_between(x_fit, y_fit - yerr_fit, y_fit + yerr_fit, **fit_err_par)
            ax.set_ylim(ymin, ymax)  # fits can sometimes explode

            # residuals
            res = y - np.interp(x, x_fit, y_fit)
            ax2.errorbar(x, res, yerr, **data_par)
            ymin, ymax = ax2.get_ylim()
            ax2.plot(x_fit, np.zeros_like(y_fit), **fit_par)
            ax2.fill_between(x_fit, -yerr_fit, yerr_fit, **fit_err_par)
            lim = np.max([np.abs(ymin), ymax])
            ax2.set_ylim(-lim, lim)  # symmetric limits

            for axis in [ax, ax2]:
                axis.axvline(
                    x=tmax - t_offset, color="k", linestyle="--", alpha=0.4
                )
                axis.xaxis.set_tick_params(labelsize=15)
                axis.yaxis.set_tick_params(labelsize=15)
                axis.minorticks_on()
                axis.tick_params(
                    which="major",
                    length=6,
                    width=1,
                    direction="in",
                    top=True,
                    right=True,
                )
                axis.tick_params(
                    which="minor",
                    length=3,
                    width=1,
                    direction="in",
                    top=True,
                    right=True,
                )
            ax.set_xticks([])
            ax.legend(fontsize=16)

            ax.set_xlim(x_fit.min(), x_fit.max())
            ax2.set_xlim(x_fit.min(), x_fit.max())

        fig.text(
            0.5,
            0.92,
            f"{self.name} (z = {self.z:.5})",
            ha="center",
            fontsize=20,
            family="serif",
        )
        fig.text(
            0.5,
            0.05,
            f"Time - {t_offset} [days]",
            ha="center",
            fontsize=18,
            family="serif",
        )
        if not plot_mag:
            fig.text(
                0.05,
                0.5,
                f"Flux (ZP = {ZP})",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )
        else:
            fig.text(
                0.05,
                0.5,
                r"Apparent Magnitude",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )

        if fig_name is not None:
            plt.savefig(fig_name)

        plt.show()

    def plot_rest_fits(self, plot_mag=False, fig_name=None):
        """Plots the rest-frame light-curves fits results.

        Plots the observed data for each band together with the gaussian process fits.
        The time of :math:`B`-band peak is shown as a vertical dashed line.

        Parameters
        ----------
        plot_mag : bool, default ``False``
            If ``True``, plots the bands in magnitude space.
        fig_name : str, default ``None``
            If  given, name of the output plot.
        """
        palette1 = [plt.get_cmap("Dark2")(i) for i in np.arange(8)]
        palette2 = [plt.get_cmap("Set1")(i) for i in np.arange(9)]
        palette3 = [plt.get_cmap("Pastel2")(i) for i in np.arange(8)]
        colours = palette1 + palette2 + palette3 + palette1 + palette2

        # shift in time for visualization purposes
        try:
            tmax_str = str(self.tmax.astype(int))
        except:
            tmax_str = str(self.tmax.astype(int))
        zeros = "0" * len(tmax_str[2:])
        t_offset = int(tmax_str[:2] + zeros)

        data = self.rest_lcs
        ZP = 27.5  # global zero-point for visualization

        h = 3  # columns
        v = math.ceil(len(data.bands) / h)  # rows

        fig = plt.figure(figsize=(15, 5 * v))
        gs = gridspec.GridSpec(v, h)

        for i, band in enumerate(data.bands):
            j = math.ceil(i % h)
            k = i // h
            ax = plt.subplot(gs[k, j])

            x = data[band].times - t_offset
            if plot_mag is False:
                y_norm = change_zp(1.0, data[band].zp, ZP)
                y = data[band].fluxes * y_norm
                yerr = data[band].flux_errors * y_norm
            else:
                y = data[band].magnitudes
                yerr = data[band].mag_errors
                
            colour = colours[i]
            fit_par = dict(ls="-", lw=2, zorder=16, color=colour)
            fit_err_par = dict(alpha=0.3, color=colour)

            # light curves
            ax.plot(x, y, label=band, **fit_par)
            ax.fill_between(x, y - yerr, y + yerr, **fit_err_par)
            ymin, ymax = y.min(), y.max()
            ax.set_ylim(ymin, ymax)  # fits can sometimes explode
            if plot_mag is True:
                ax.invert_yaxis()

            ax.xaxis.set_tick_params(labelsize=15)
            ax.yaxis.set_tick_params(labelsize=15)
            ax.minorticks_on()
            ax.tick_params(
                which="major",
                length=6,
                width=1,
                direction="in",
                top=True,
                right=True,
            )
            ax.tick_params(
                which="minor",
                length=3,
                width=1,
                direction="in",
                top=True,
                right=True,
            )
            ax.legend(fontsize=16)
            ax.set_xlim(x.min(), x.max())

        fig.text(
            0.5,
            0.92,
            f"{self.name} (z = {self.z:.5})",
            ha="center",
            fontsize=20,
            family="serif",
        )
        fig.text(
            0.5,
            0.05,
            f"Time - {t_offset} [days]",
            ha="center",
            fontsize=18,
            family="serif",
        )
        if not plot_mag:
            fig.text(
                0.05,
                0.5,
                f"Flux (ZP = {ZP})",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )
        else:
            fig.text(
                0.05,
                0.5,
                r"Apparent Magnitude",
                va="center",
                rotation="vertical",
                fontsize=18,
                family="serif",
            )

        if fig_name is not None:
            plt.savefig(fig_name)

        plt.show()

    def export_fits(self, output_file=None):
        """Exports the light-curve fits into an output file.

        Parameters
        ----------
        output_file : str, default ``None``
            Name of the output file. If ``None``, the output
            file is of the form ``<SN name>_fits.dat``.
        """
        if output_file is None:
            output_file = f"{self.name}_fits.dat"

        df_list = []
        columns = ["times", "fluxes", "flux_errors", "magnitudes", "mag_errors", "zp", "band"]

        for band in self.bands:
            band_info = self.lc_fits[band]
            band_dict = {key: band_info[key] for key in columns}
            band_df = pd.DataFrame(band_dict)

            rounding_dict = {key: 3 for key in ["times", "magnitudes", "mag_errors", "zp"]}
            band_df = band_df.round(rounding_dict)
            df_list.append(band_df[columns])

        df_fits = pd.concat(df_list)
        df_fits.to_csv(output_file, sep="\t", index=False)

    def export_restframe_lcs(self, output_file=None):
        """Exports the corrected, rest-frame light-curves into an output file.

        Parameters
        ----------
        output_file : str, default ``None``
            Name of the output file. If ``None``, the output
        file is of the form ``<SN name>_restframe_lcs.dat``.
        """

        if output_file is None:
            output_file = f"{self.name}_restframe_lcs.dat"

        df_list = []
        columns = ["times", "fluxes", "flux_errors", "magnitudes", "mag_errors", "zp", "band"]

        for band in self.rest_lcs.bands:
            band_info = self.rest_lcs[band]
            band_dict = {key: band_info[key] for key in columns}
            band_df = pd.DataFrame(band_dict)

            rounding_dict = {key: 3 for key in ["times", "magnitudes", "mag_errors", "zp"]}
            band_df = band_df.round(rounding_dict)
            df_list.append(band_df[columns])

        df_fits = pd.concat(df_list)
        df_fits.to_csv(output_file, sep="\t", index=False)



import os
import unittest
import piscola
import matplotlib.pyplot as plt

import warnings


class TestPiscola(unittest.TestCase):
    def test_plot(self):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            sn = piscola.call_sn("data/03D1au.dat")
            plt.ion()
            sn.plot_lcs()

    def test_fit_and_outputs(self):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            sn = piscola.call_sn("data/03D1au.dat")
            sn.fit()

            plt.ion()
            sn.plot_fits()

            sn.export_fits()
            sn.export_restframe_lcs()

    def test_save_and_load(self):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            sn = piscola.call_sn("data/03D1au.dat")
            sn.save()

            outfile = "03D1au.pisco"
            _ = piscola.load_sn(outfile)
            os.remove(outfile)  # clean directory

if __name__ == "__main__":
    unittest.main()



import numpy as np
from astropy.stats import mad_std, sigma_clip

def flux_to_mag(flux, flux_err=np.array([0.0]), zp=0.0):
    """Converts fluxes to magnitudes, propagating errors if given.

    Note: if there are negative or zero fluxes, these are converted to NaN values.

    Parameters
    ----------
    flux : array
        Array of fluxes.
    flux_err : array, default ``np.array([0.0])``
        Array of flux errors.
    zp : float or array, default ``0.0``
        Zero points.

    Returns
    -------
    mag : array
        Fluxes converted to magnitudes.
    mag_err : array
        Flux errors converted to errors in magnitudes.
    """
    flux_ = flux.copy()
    if isinstance(flux_, np.ndarray):
        flux_[flux_ <= 0.0] = np.nan

    flux_err_ = flux_err.copy()
    if isinstance(flux_err_, np.ndarray):
        flux_err_[flux_err_ <= 0.0] = np.nan

    mag = -2.5 * np.log10(flux_) + zp
    mag_err = np.abs(2.5 * flux_err_ / (flux_ * np.log(10)))

    return mag, mag_err


def mag_to_flux(mag, mag_err=np.array([0.0]), zp=0.0):
    """Converts magnitudes to fluxes, propagating errors if given.

    Parameters
    ----------
    mag : array
        Array of magnitudes.
    mag_err : array, default ``np.array([0.0])``
        Array of magnitude errors.
    zp : float or array, default ``0.0``
        Zero points.

    Returns
    -------
    flux : array
        Magnitudes converted to fluxes.
    flux_err : array
        Magnitude errors converted to errors in fluxes.
    """
    mag_ = mag.copy()
    if isinstance(mag_, np.ndarray):
        mag_[mag_ <= 0.0] = np.nan

    mag_err_ = mag_err.copy()
    if isinstance(mag_err_, np.ndarray):
        mag_err_[mag_err_ <= 0.0] = np.nan

    flux = 10 ** (-0.4 * (mag_ - zp))
    flux_err = np.abs(flux * 0.4 * np.log(10) * mag_err_)

    return flux, flux_err

def calculate_robust_stats(values, sigma=5):
    """Calculates a robust mean and standard deviation not sensitive to outliers.

    Parameters
    ----------
    values : array
        Array of values.
    sigma : float, default ``5``
        Number of sigmas used for sigma clipping.

    Returns
    -------
    robust_mean : float
        Robust mean.
    robust_std : float
        Robust standard deviation.
    robust_mask: bool array
        Mask with valid values as ``True``.
    """
    std = mad_std(values, ignore_nan=True)    # MAD std that "avoids" outliers    
    robust_values = sigma_clip(values, sigma=sigma, sigma_lower=std, sigma_upper=std, maxiters=1)
    robust_mean = np.mean(robust_values)
    robust_std = np.std(robust_values)
    robust_mask = ~robust_values.mask

    return robust_mean, robust_std, robust_mask


def change_zp(flux, zp, new_zp):
    """Converts flux units given a new zero-point.

    **Note:** this assumes that the new zero-point is in the same magnitude system as the current one.

    Parameters
    ----------
    flux : float or array
        Fluxes.
    zp : float or array
        Current zero-point for the given fluxes.
    new_zp : float or array
        New zero-point to convert the flux units.

    Returns
    -------
    new_flux : float or array
        Fluxes with with a new zero-point.
    """
    new_flux = flux * 10 ** (-0.4 * (zp - new_zp))

    return new_flux

def _integrate_filter(
    sed_wave, sed_flux, filter_wave, filter_transmission, response_type="photon"
):
    """Calculates the flux density of a spectral energy distribution (SED)
    given a filter transmission function.

    Mainly used by the :func:`extinction_correction` module.

    Parameters
    ----------
    sed_wave : array
        SED's wavelength range.
    sed_flux : array
        SED's flux density distribution.
    filter_wave : array
        Filter's wavelength range.
    filter_transmission : array
        Filter's transmission function.
    response_type : str, default ``photon``
        Filter's response type. Either ``photon`` or ``energy``.

    Returns
    -------
    flux_filter : float
        Flux density.
    """

    blue_edge_covered = sed_wave.min() <= filter_wave.min()
    red_edge_covered = sed_wave.max() >= filter_wave.max()
    err_message = "The SED does not completely overlap with {self.band} filter."
    assert blue_edge_covered and red_edge_covered, err_message

    transmission = filter_transmission.copy()
    # check filter response type
    if response_type == "energy":
        transmission /= filter_wave

    transmission = np.interp(sed_wave, filter_wave, transmission, left=0.0, right=0.0)
    I1 = np.trapz(sed_flux * transmission * sed_wave, sed_wave)
    I2 = np.trapz(filter_transmission * filter_wave, filter_wave)
    flux_filter = I1 / I2

    return flux_filter


import os
import tarfile
import requests
import numpy as np
import matplotlib.pyplot as plt

from sfdmap2 import sfdmap
import extinction

import piscola
from .utils import _integrate_filter

pisco_path = piscola.__path__[0]
dustmaps_dir = os.path.join(pisco_path, "sfddata-master")


def _download_dustmaps():
    """Downloads the dust maps for extinction calculation if they are not found
    locally.
    """
    global dustmaps_dir

    # check if files already exist locally
    dust_files = [
        os.path.join(dustmaps_dir, f"SFD_dust_4096_{sky}gp.fits") for sky in ["n", "s"]
    ]
    mask_files = [
        os.path.join(dustmaps_dir, f"SFD_mask_4096_{sky}gp.fits") for sky in ["n", "s"]
    ]
    maps_files = dust_files + mask_files
    existing_files = [os.path.isfile(file) for file in maps_files]

    if all(existing_files) != True:
        # download dust maps
        sfdmaps_url = "https://github.com/kbarbary/sfddata/archive/master.tar.gz"
        response = requests.get(sfdmaps_url)

        master_tar = "master.tar.gz"
        with open(master_tar, "wb") as file:
            file.write(response.content)

        # extract tar file under mapsdir directory
        tar = tarfile.open(master_tar)
        tar.extractall(pisco_path)
        tar.close()
        os.remove(master_tar)


def redden(
    wave, flux, ra, dec, scaling=0.86, reddening_law="fitzpatrick99", r_v=3.1, ebv=None
):
    """Reddens the given spectrum, given a right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    wave : array
        Wavelength values.
    flux : array
        Flux density values.
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: str, default ``fitzpatrick99``
        Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
        ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
        :math:`R_V = 3.1`.)
    r_v : float, default ``3.1``
        Total-to-selective extinction ratio (:math:`R_V`)
    ebv : float, default ``None``
        Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

    Returns
    -------
    redden_flux : array
        Redden flux values.

    """
    _download_dustmaps()
    global dustmaps_dir

    if ebv is None:
        dustmaps_dir = os.path.join(pisco_path, "sfddata-master")
        m = sfdmap.SFDMap(mapdir=dustmaps_dir, scaling=scaling)
        ebv = m.ebv(ra, dec)  # RA and DEC in degrees

    a_v = r_v * ebv

    rl_list = ["ccm89", "odonnell94", "fitzpatrick99", "calzetti00", "fm07"]
    err_message = f"Choose one of the available reddening laws: {rl_list}"
    assert reddening_law in rl_list, err_message

    wave = np.copy(wave.astype("double"))
    if reddening_law == "ccm89":
        ext = extinction.ccm89(wave, a_v, r_v)
    elif reddening_law == "odonnell94":
        ext = extinction.odonnell94(wave, a_v, r_v)
    elif reddening_law == "fitzpatrick99":
        ext = extinction.fitzpatrick99(wave, a_v, r_v)
    elif reddening_law == "calzetti00":
        ext = extinction.calzetti00(wave, a_v, r_v)
    elif reddening_law == "fm07":
        ext = extinction.fm07(wave, a_v)

    redden_flux = extinction.apply(ext, flux)

    return redden_flux


def deredden(
    wave, flux, ra, dec, scaling=0.86, reddening_law="fitzpatrick99", r_v=3.1, ebv=None
):
    """Dereddens the given spectrum, given a right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    wave : array
        Wavelength values.
    flux : array
        Flux density values.
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: str, default ``fitzpatrick99``
        Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
        ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
        :math:`R_V = 3.1`.)
    r_v : float, default ``3.1``
        Total-to-selective extinction ratio (:math:`R_V`)
    ebv : float, default ``None``
        Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

    Returns
    -------
    deredden_flux : array
        Deredden flux values.

    """
    _download_dustmaps()
    global dustmaps_dir

    if ebv is None:
        m = sfdmap.SFDMap(mapdir=dustmaps_dir, scaling=scaling)
        ebv = m.ebv(ra, dec)  # RA and DEC in degrees

    a_v = r_v * ebv

    rl_list = ["ccm89", "odonnell94", "fitzpatrick99", "calzetti00", "fm07"]
    assert (
        reddening_law in rl_list
    ), f"Choose one of the available reddening laws: {rl_list}"

    if reddening_law == "ccm89":
        ext = extinction.ccm89(wave, a_v, r_v)
    elif reddening_law == "odonnell94":
        ext = extinction.odonnell94(wave, a_v, r_v)
    elif reddening_law == "fitzpatrick99":
        ext = extinction.fitzpatrick99(wave, a_v, r_v)
    elif reddening_law == "calzetti00":
        ext = extinction.calzetti00(wave, a_v, r_v)
    elif reddening_law == "fm07":
        ext = extinction.fm07(wave, a_v)

    deredden_flux = extinction.remove(ext, flux)

    return deredden_flux


def calculate_ebv(ra, dec, scaling=0.86):
    """Calculates Milky Way colour excess, :math:`E(B-V)`, for a given right ascension and declination.

    Parameters
    ----------
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Finkbeiner & Davis (1998).

    Returns
    -------
    ebv :  float
        Reddening value, :math:`E(B-V)``.
    """
    _download_dustmaps()
    global dustmaps_dir
    m = sfdmap.SFDMap(mapdir=dustmaps_dir, scaling=scaling)
    ebv = m.ebv(ra, dec)  # RA and DEC in degrees

    return ebv


def extinction_filter(
    filter_wave,
    filter_response,
    ra,
    dec,
    scaling=0.86,
    reddening_law="fitzpatrick99",
    r_v=3.1,
    ebv=None,
):
    """Calculates the extinction for a given filter, given a right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    filter_wave : array
        Filter's wavelength range.
    filter_response : array
        Filter's response function.
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: str, default ``fitzpatrick99``
        Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
        ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
        :math:`R_V = 3.1`.)
    r_v : float, default ``3.1``
        Total-to-selective extinction ratio (:math:`R_V`)
    ebv : float, default ``None``
        Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

    Returns
    -------
    A : float
        Extinction value in magnitudes.
    """
    flux = 100
    deredden_flux = deredden(
        filter_wave, flux, ra, dec, scaling, reddening_law, r_v, ebv
    )

    f1 = _integrate_filter(filter_wave, flux, filter_wave, filter_response)
    f2 = _integrate_filter(filter_wave, deredden_flux, filter_wave, filter_response)
    A = -2.5 * np.log10(f1 / f2)

    return A

def extinction_at_wavelength(
    wave,
    ra,
    dec,
    scaling=0.86,
    reddening_law="fitzpatrick99",
    r_v=3.1,
    ebv=None,
):
    """Calculates the extinction at a given wavelength, given a right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    wave : float
        Wavelength in agnstroms.
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: str, default ``fitzpatrick99``
        Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
        ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
        :math:`R_V = 3.1`.)
    r_v : float, default ``3.1``
        Total-to-selective extinction ratio (:math:`R_V`)
    ebv : float, default ``None``
        Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

    Returns
    -------
    A : float
        Extinction value in magnitudes.
    """
    wave_array = np.array([wave - 1, wave, wave + 1])
    resp_array = np.array([1.0, 1.0, 1.0])
    A = extinction_filter(wave_array, resp_array, ra, dec, 
                          scaling, reddening_law, r_v, ebv)

    return A


def extinction_curve(
    ra, dec, scaling=0.86, reddening_law="fitzpatrick99", r_v=3.1, ebv=None
):
    """Plots the extinction curve for a given right ascension and declination or :math:`E(B-V)`.

    Parameters
    ----------
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    scaling: float, default ``0.86``
        Calibration of the Milky Way dust maps. Either ``0.86``
        for the Schlafly & Finkbeiner (2011) recalibration or ``1.0`` for the original
        dust map of Schlegel, Fikbeiner & Davis (1998).
    reddening_law: str, default ``fitzpatrick99``
        Reddening law. The options are: ``ccm89`` (Cardelli, Clayton & Mathis 1989), ``odonnell94`` (O’Donnell 1994),
        ``fitzpatrick99`` (Fitzpatrick 1999), ``calzetti00`` (Calzetti 2000) and ``fm07`` (Fitzpatrick & Massa 2007 with
        :math:`R_V = 3.1`.)
    r_v : float, default ``3.1``
        Total-to-selective extinction ratio (:math:`R_V`)
    ebv : float, default ``None``
        Colour excess (:math:`E(B-V)`). If given, this is used instead of the dust map value.

    """
    flux = 100
    wave = np.arange(1000, 25001)  # in Angstroms
    deredden_flux = deredden(wave, flux, ra, dec, scaling, reddening_law, r_v, ebv)
    ff = 1 - flux / deredden_flux

    f, ax = plt.subplots(figsize=(8, 6))
    ax.plot(wave, ff)

    ax.set_xlabel(r"wavelength ($\AA$)", fontsize=18)
    ax.set_ylabel("fraction of extinction", fontsize=18)
    ax.set_title(r"Extinction curve", fontsize=18)
    ax.xaxis.set_tick_params(labelsize=15)
    ax.yaxis.set_tick_params(labelsize=15)

    plt.show()



# -*- coding: utf-8 -*-

from .sn_class import call_sn, load_sn, Supernova
from . import filters_class

from . import extinction_correction
from . import gaussian_process
from . import utils
from ._version import __version__



import unittest
import os
import numpy as np
import matplotlib.pyplot as plt
import piscola
from piscola.extinction_correction import (
    calculate_ebv,
    redden,
    deredden,
    extinction_filter,
    extinction_curve,
)

import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)


class TestPiscola(unittest.TestCase):
    def test_calculate_ebv(self):
        ra, dec = 30, 30

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            ebv_SF2011 = calculate_ebv(ra, dec, scaling=0.86)
            ebv_SFD1998 = calculate_ebv(ra, dec, scaling=1)

        # reference values from https://irsa.ipac.caltech.edu/cgi-bin/bgTools/nph-bgExec
        ref_ebv_SF2011 = 0.0431  # mean value S & F (2011)
        ref_ebv_SFD1998 = 0.0502  # mean value SFD (1998)

        err_msg = (
            "Difference in E(B-V) between calculated and reference value is too large."
        )
        np.testing.assert_allclose(
            ebv_SF2011, ref_ebv_SF2011, rtol=0.02, err_msg=err_msg
        )
        np.testing.assert_allclose(
            ebv_SFD1998, ref_ebv_SFD1998, rtol=0.02, err_msg=err_msg
        )

    def test_reddening_laws(self):
        ra, dec = 30, 30

        wave = np.arange(1000, 20000)
        flux = np.ones_like(wave)

        for reddening_law in [
            "fitzpatrick99",
            "ccm89",
            "odonnell94",
            "calzetti00",
            "fm07",
        ]:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=DeprecationWarning)
                reddened_flux = redden(wave, flux, ra, dec, reddening_law=reddening_law)
            self.assertTrue(reddened_flux.mean() < flux.mean())

    def test_ebv(self):
        ra, dec = 30, 30

        wave = np.arange(1000, 20000)
        flux = np.ones_like(wave)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            reddened_flux_maps = redden(wave, flux, ra, dec)

            ebv = calculate_ebv(ra, dec)
            reddened_flux_ebv = redden(wave, flux, ra, dec, ebv=ebv)

        self.assertTrue(reddened_flux_maps.mean() == reddened_flux_ebv.mean())

    def test_deredden(self):
        ra, dec = 30, 30

        wave = np.arange(1000, 20000)
        flux = np.ones_like(wave)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            reddened_flux = redden(wave, flux, ra, dec)
            dereddened_flux = deredden(wave, reddened_flux, ra, dec)

        self.assertTrue(dereddened_flux.mean() == flux.mean())

    def test_visual_extinction(self):
        pisco_path = piscola.__path__[0]
        ra, dec = 30, 30

        filter_file = os.path.join(pisco_path, "filters", "Bessell", "Bessell_V.dat")
        filter_wave, filter_response = np.loadtxt(filter_file).T
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            A_V = extinction_filter(filter_wave, filter_response, ra, dec, r_v=3.1)
        ref_A_V = (
            0.1368  # from https://irsa.ipac.caltech.edu/cgi-bin/bgTools/nph-bgExec
        )

        err_msg = (
            "Difference in A_V between calculated and reference value is too large."
        )
        np.testing.assert_allclose(A_V, ref_A_V, rtol=0.03, err_msg=err_msg)

    def test_extinction_curve(self):
        ra, dec = 30, 30
        plt.ion()
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            extinction_curve(ra, dec)


if __name__ == "__main__":
    unittest.main()
